/*
MotorTask.h

Copyright 2001 Larry Barello

Author: Larry Barello
        larry@barello.net
*/
#define TRUE 1
#define FALSE 0
/*
  RC_MIN/MAX are the min and max timing pulse width
  to be scaled to 0-255 for controlling the PWM.

  The timing is in 8uS increments.  The default, initial
  values are for a 600uS range, or 1.2mS to 1.8ms.  Divide
  those numbers by 8uS to get the counts, below.

  20010828 - expanded range by one count so scaling could
  produce a "zero" value for PWM calculations

  The values are stored in EEPROM so a future code warrier
  can add a menu or other calibration item to update
  these values in the field.
*/
#define RC_MIN 149              // ((unsigned char)(1.2*1000/8))
#define RC_MAX 225              // ((unsigned char)(1.8*1000/8))
#define AUXLOW_THRESHOLD (0xFF*1/3)
#define AUXHIGH_THRESHOLD (0xFF*2/3) // Normalized 0-FF (1.0-2.0ms)
#define DEADBAND_RANGE 8

typedef struct
{
    unsigned char start;        // Start time   : Keep this order
    unsigned char end;          // end time     : Keep this order
    unsigned char done;         // pulse flags  : Keep this order
    unsigned char low;          // Scaling factor
    unsigned char high;         // scaling factor
    unsigned char valid;        // Marks data as valid or not.
    unsigned char pulse;        // count of pulses received
    unsigned char rcpulse;      // last calculated R/C pulse width
    unsigned char width;        // Normalized pulse width
    char          Drive;        // Temp storage for LCD display
}
RCcontrolBlock, *pRCcontrolBlock;


typedef union
{
    struct
    {
        unsigned char invalid_mode:1;   // Bad value in PWM mode
        unsigned char invalid_rate:1;   // Bad value in frequency selec
        unsigned char invalid_mix:1;    // Bad mixing mode
        unsigned char invalid_rc0:1;
        unsigned char invalid_rc1:1;    // Invalid R/C pulse kill bits
        unsigned char invalid_rc2:1;
        unsigned char lcd_disable:1;    // User interface kill bit
    };
    unsigned char Output;
}
DisableFlags, *pDisableFlags;

#define DEFAULT_DISABLE 0x40            // Default is lcd_disable = TRUE

extern RCcontrolBlock RC0, RC1, RC2;
extern DisableFlags fDisable;
extern  char PWM_Mode,
             PWM_Rate,
             LeftSupply,
             RightSupply,
             MixedMode,
             DeadBand,
             SlewRate,
             BrakeMode;

