#include "avrx.inc"
;
; RCpulse.s
;
; STACK: 4
; INTERRUPT Latency: ~16 cycles
;
; Assembly code for the RC pulse measuring interrupts.  These run
; outside of AvrX and are carefully coded to use the minimum stack
; space and minimum interrupt latency.  Each can interrupt the other.
;
; The GCC coded versions are about twice as slow and use about twice 
; as much stack (v3.0).  This is an issue since each task stack needs
; to have enough room to accomodate any usage by these routines.
;
; Offsets into the RC data structure (MotorTask.h)
;
#define RC_START 0	// These are all bytes in the struct.
#define RC_END	1
#define RC_DONE 2


	.global	_interrupt0_
	.func	_interrupt0_
_interrupt0_:
	push	R16
	in	R16, SREG	; Stash context
	push	R16

	in	R16, MCUCR	; Determine rising/falling edge
	sbrs	R16, ISC00
	rjmp	1f

	cbr	R16, BV(ISC00)	; Rising, switch to fall
	out	MCUCR, R16
	in	R16, TCNT0
	sts	RC0+RC_START, R16; Stash the start time.
	rjmp	2f
1:
	sbr	R16, BV(ISC00)	; Falling, switch to rising
	out	MCUCR, R16
	in	R16, TCNT0
	sts	RC0+RC_END, R16	; Stash the end time
	ldi	R16, 1
	sts	RC0+RC_DONE, R16
2:
	pop	R16		; Restore context
	out	SREG, R16
	pop	R16
	reti
	.endfunc

	.global	_interrupt1_
	.func	_interrupt1_
_interrupt1_:
	push	R16
	in	R16, SREG
	push	R16

	in	R16, MCUCR
	sbrs	R16, ISC10
	rjmp	1f

	cbr	R16, BV(ISC10)
	out	MCUCR, R16
	in	R16, TCNT0
	sts	RC1+RC_START, R16
	rjmp	2f
1:
	sbr	R16, BV(ISC10)
	out	MCUCR, R16
	in	R16, TCNT0
	sts	RC1+RC_END, R16
	ldi	R16, 1
	sts	RC1+RC_DONE, R16
2:
	pop	R16
	out	SREG, R16
	pop	R16
	reti
	.endfunc

;
; _comparator_
;
; This routine does the same function as the above routines
; with the exception that it uses the ADC as a third edge 
; selectable interrupt input.  Cool.
;
; Because direct bit manipulation works on the ACSR, we don't
; even need to preserve the SREG!
;
	.func	_comparator_
	.global	_comparator_
_comparator_:
	push	R16
	in	R16, TCNT0

	sbis	ACSR, ACIS0
	rjmp	1f

	cbi	ACSR, ACIS0
	sts	RC2+RC_START, R16
	pop	R16
	reti
1:
	sbi	ACSR, ACIS0
	sts	RC2+RC_END, R16
	ldi	R16, 1
	sts	RC2+RC_DONE, R16

	pop	R16
	reti

	.endfunc
